"""utility functions:

   - tests for common types
   - simple format of h:m:s from seconds
"""

from __future__ import print_function, absolute_import
from six import string_types, integer_types

def get_filename_using_tk():
    """Choose a filename using Tk"""
    import tkinter as tk
    root = tk.Tk()
    root.withdraw()
    fname = tk.filedialog.askopenfilename(filetypes = [("TSV","*.tsv")])
    print('Selected file {}'.format(fname))
    return fname


def _is_sequence(arg):
    isstring = isinstance(arg, string_types)
    isothersequences = hasattr(arg, "__getitem__") or hasattr(arg, "__iter__")
    return not isstring and isothersequences


def _is_string(a):
    return isinstance(a, string_types)


def _is_number(a):
    return isinstance(a, float) or isinstance(a, integer_types)


def s2HMS(seconds):
    m, s = divmod(seconds, 60.0)
    h, m = divmod(m, 60.0)
    if h == 0:
        return "%02dm %06.3fs" % (m, s)
    return "%dh %02dm %06.3fs" % (h, m, s)



from __future__ import print_function

from collections import OrderedDict
from numpy import nan

import pandas as pd

from metabolinks.spectra import AlignedSpectra, read_spectra_from_xcel
from metabolinks.similarity import compute_similarity_measures

def concat_peaks(spectra, verbose=True):
    if verbose:
        print ('- Joining data...')

    n = len(spectra)
    dfs = []
    # tag with sample names, concat vertically and sort by m/z
    for spectrum in spectra:
        # create DataFrame with columns 'm/z', 'I', and '_sample'
        mzcol = pd.Series(spectrum.mz, index= range(len(spectrum.data)), 
                          name='m/z')
        intensity_column = pd.Series(spectrum.data.iloc[:, 0].values,
                                     index=mzcol.index,
                                     name='I')
        labelcol = pd.Series(spectrum.sample_name, index=mzcol.index,
                             name='_sample')

        dfs.append(pd.concat([mzcol, intensity_column, labelcol], axis=1))
    
    # sort by m/z
    cdf = pd.concat(dfs)
    cdf = cdf.sort_values(by='m/z')
    #reindex with increasing integers
    cdf.index = list(range(len(cdf)))

    if verbose:
        print('  done, {} peaks in {} samples'.format(cdf.shape[0], n))
    return cdf


def are_near(d1, d2, ppmtol):
    """Predicate: flags if two entries should belong to the same compound"""
    reltol = ppmtol * 1.0e-6
    # two consecutive peaks from the same sample
    # should not belong to the same group
    if d1['_sample'] == d2['_sample']:
        return False
    m1, m2 = d1['m/z'], d2['m/z']
    if (m2 - m1) / m2 <= reltol:
        return True
    return False


def group_peaks(df, sample_ids, labels=None,
                ppmtol=1.0, min_samples=1,
                verbose=True):
    """Group peaks from different samples.

       Peaks are grouped if their relative mass difference is below
       a mass difference tolerance, in ppm."""

    if verbose:
        print ('- Aligning...')
    glabel = 0
    start = 0
    glabels = [0]

    for i in range(len(df)):
        if i == start:
            continue
        d1 = df.iloc[start]
        d2 = df.iloc[i]
        assert isinstance(d1['m/z'], (int, float))
        assert isinstance(d2['m/z'], (int, float))

        if not are_near(d1, d2, ppmtol):
            glabel += 1
            start = i

        glabels.append(glabel)

    result = pd.concat([df, pd.Series(glabels, index=df.index, name='_group')],
                        axis=1)
    grouped = result.groupby('_group')

    colnames = ['m/z'] + sample_ids + ['#samples', 'range_ppm']
    intensities = {sname: list() for sname in colnames}

    for gname, gvalue in grouped:
        tagged = gvalue.set_index('_sample')

        intensities['m/z'].append(tagged['m/z'].mean())
        intensities['#samples'].append(len(tagged))
        m_min = tagged['m/z'].min()
        m_max = tagged['m/z'].max()
        range_ppm = 1e6 * (m_max - m_min) / m_max
        intensities['range_ppm'].append(range_ppm)

        for s in sample_ids:
            i = int(tagged.loc[s, 'I']) if s in tagged.index else nan
            intensities[s].append(i)

    result = pd.DataFrame(intensities, columns=colnames)
    n_non_discarded = len(result)

    if verbose:
        print('  done, {} aligned peaks'.format(n_non_discarded))

    if min_samples > 1:
        result = result[result['#samples'] >= min_samples]

    if verbose:
        if min_samples > 1:
            n_discarded = n_non_discarded - len(result)
            print('- {} peaks were discarded (#samples < {})'.format(n_discarded, min_samples))

        counts_table = result['#samples'].value_counts().sort_index()
        for n, c in counts_table.iteritems():
            print ('{:5d} peaks in {} samples'.format(c, n))
        print('  {:3d} total'.format(len(result)))

    result = result.set_index('m/z')
    res = AlignedSpectra(result, sample_names=sample_ids, labels=labels)

    return res


def align_spectra(spectra,
                  ppmtol=1.0,
                  min_samples=1,
                  fillna=None,
                  verbose=True):

    samplenames = [s.sample_name for s in spectra]
    no_labels = False
    for s in spectra:
        if s.label is None:
            no_labels = True
            break
    if no_labels:
        labels = None
    else:
        labels = [s.label for s in spectra]
    
    if verbose:
        print ('  Sample names:', samplenames)
        if not no_labels:
            print ('  Labels:', labels)
            
    mdf = concat_peaks(spectra, verbose=verbose)
    gdf = group_peaks(mdf, samplenames, labels=labels,
                      ppmtol=ppmtol,
                      min_samples=min_samples,
                      verbose=verbose)
    
    #gdf.compute_similarity_measures()
    
    if fillna is not None:
        gdf = gdf.fillna(fillna)
    return gdf


def save_aligned_to_excel(fname, aligned_dict):
    writer = pd.ExcelWriter(fname, engine='xlsxwriter')

    for sname in aligned_dict:
        
        aligned_spectra = aligned_dict[sname]
        
        sim = compute_similarity_measures(aligned_spectra)
                
        sample_similarity = sim.sample_similarity
        label_similarity = sim.label_similarity
        unique_labels = sim.unique_labels
                
        results = aligned_spectra.data
        results = results.reset_index(level=0)
        
        n_compounds, ncols = results.shape

        # write Pandas DataFrame
        results.to_excel(writer, sheet_name=sname,
                         index=False, startrow=1, startcol=1)

        workbook = writer.book
        sh = writer.sheets[sname]

        # center '#samples' column
        format = workbook.add_format({'align': 'center'})
        sh.set_column(ncols-1, ncols-1, None, format)

        # change displayed precision in 'm/z' column
        format2 = workbook.add_format({'num_format': '0.0000000'})
        sh.set_column(1, 1, 15, format2)

        # change widths
        sh.set_column(2, ncols - 2, 25)
        #sh.set_column(1, 1, 15)
        format3 = workbook.add_format({'num_format': '0.0000'})
        sh.set_column(ncols, ncols, 15, format3)
        
        # Create report sheet
        rsheetname = sname + ' report'
        sh = workbook.add_worksheet(rsheetname)
        
        row_offset = 1
        
        sh.write_string(row_offset, 1, 'Peak reproducibility')
        
        row_offset = row_offset + 2
        
        # create table of replica counts
        sh.write_string(row_offset, 2, '#samples')
        counts = results['#samples'].value_counts().sort_index()
        sh.write_column(row_offset + 1, 2, counts.index)
        sh.write_column(row_offset + 1, 3, counts.values)

        sh.write_string(row_offset + 1 + len(counts), 2, 'total')
        sh.write_number(row_offset + 1 + len(counts), 3, len(results))

        # Add pie chart of replica counts
        chart = workbook.add_chart({'type': 'pie'})
        chart.add_series({'values': [rsheetname, row_offset+1, 3, row_offset + len(counts), 3],
                          'categories': [rsheetname, row_offset+1, 2, row_offset + len(counts), 2],
                          'name': '#samples'})
        chart.set_size({'width': 380, 'height': 288})
        sh.insert_chart(1, 5, chart)

        row_offset = row_offset + len(counts) + 7

        sh.write_string(row_offset, 1, 'Sample sizes')
        
        row_offset = row_offset + 2
        
        sh.write_row(row_offset, 1, aligned_spectra.sample_names)
        for i in range(sample_similarity.shape[0]):
            sh.write_number(row_offset + 1, 1 + i, sample_similarity[i, i])

        row_offset = row_offset + 4
        
        sh.write_string(row_offset, 1, 'Sample similarity')
        
        row_offset = row_offset + 2

        sh.write_row(row_offset, 2, aligned_spectra.sample_names)
        sh.write_column(row_offset + 1, 1, aligned_spectra.sample_names)
        for i in range(sample_similarity.shape[0]):
            sh.write_row(row_offset + i + 1, 2, sample_similarity[i, :])
            
        if label_similarity is not None:
            row_offset = row_offset + sample_similarity.shape[0] + 3
            
            sh.write_string(row_offset, 1, 'Label similarity')
            
            row_offset = row_offset + 2

            sh.write_row(row_offset, 2, unique_labels)
            sh.write_column(row_offset + 1, 1, unique_labels)
            for i in range(label_similarity.shape[0]):
                sh.write_row(row_offset + i + 1, 2, label_similarity[i, :])

    writer.save()
    print('Created file\n{}'.format(fname))


def align_spectra_in_excel(fname,
                           save_to_excel=None,
                           ppmtol=1.0,
                           min_samples=1,
                           sample_names=None,
                           labels=None,
                           header_row=1,
                           fillna=None,
                           verbose=True):

    spectra_table = read_spectra_from_xcel(fname,
                                           sample_names=sample_names,
                                           labels=labels,
                                           header_row=header_row,
                                           verbose=verbose)

    if verbose:
        print('\n------ Aligning spectra ------')

    aligned_spectra = OrderedDict()
    for sheetname, spectra in spectra_table.items():
        if verbose:
            print ('\n-- sheet "{}"'.format(sheetname))
        aligned_spectra[sheetname] = align_spectra(spectra,
                                              ppmtol=ppmtol,
                                              min_samples=min_samples,
                                              fillna=fillna,
                                              verbose=verbose)
    if save_to_excel is not None:
        save_aligned_to_excel(save_to_excel, aligned_spectra)
    
    return aligned_spectra


if __name__ == '__main__':
    ppmtol = 1.0
    min_samples = 1

    fname = 'data/data_to_align.xlsx'
    out_fname = 'data/aligned_data.xlsx'

    header_row = 2
    sample_names = ['S1', 'S2', 'S3']
    labels = ['wt', 'mod', 'mod']
    #sample_names = 1

    aligned = align_spectra_in_excel(fname,
                           save_to_excel=out_fname,
                           ppmtol=ppmtol,
                           min_samples=min_samples,
                           labels=labels,
                           header_row=header_row,
                           sample_names=sample_names)
        
    print('+++++++++++++++++++++++++++++++++++++++++++++++++++++')
    for sheetname in aligned:
        print(aligned[sheetname])
        print('-----------------------------------')




"""Similarity of peak lists based on peak assignments (m/z, formulae).
"""

import numpy as np

class SimilarityMeasures(object):
    """A container that holds the results of similarity measures.
    """
    
    def __init__(self):
        self.sample_similarity = None
        self.label_similarity = None
        self.unique_labels = None


def compute_similarity_measures(aligned):
    # compute counts and Jaccard index by samples
    similarities = SimilarityMeasures()
    
    n = len(aligned.sample_names)
    smatrix = np.zeros((n, n))
    for i1 in range(n-1):
        for i2 in range(i1+1, n):
            mz1 = aligned.sample(aligned.sample_names[i1]).mz
            mz2 = aligned.sample(aligned.sample_names[i2]).mz
            smatrix[i1, i1] = len(mz1)
            smatrix[i2, i2] = len(mz2)
            set1 = set(mz1)
            set2 = set(mz2)
            u12 = set1.union(set2)
            i12 = set1.intersection(set2)
            smatrix[i1, i2] = len(i12)
            jaccard = len(i12) / len(u12)
            smatrix[i2, i1] = jaccard
    similarities.sample_similarity = smatrix
    
    if aligned.labels is not None:
        # build list of unique labels
        slabels = [aligned.labels[0]]
        for i in range(1, len(aligned.labels)):
            label = aligned.labels[i]
            if label not in slabels:
                slabels.append(label)
        mzs = {}
        for label in slabels:
            mzs[label] = aligned.label(label).mz
        # compute intersection counts and Jaccard index
        n = len(slabels)
        lmatrix = np.zeros((n, n))
        for i1 in range(n-1):
            for i2 in range(i1+1, n):
                label1 = slabels[i1]
                label2 = slabels[i2]
                set1 = set(mzs[label1])
                set2 = set(mzs[label2])
                lmatrix[i1, i1] = len(set1)
                lmatrix[i2, i2] = len(set2)
                u12 = set1.union(set2)
                i12 = set1.intersection(set2)
                lmatrix[i1, i2] = len(i12)
                jaccard = len(i12) / len(u12)
                lmatrix[i2, i1] = jaccard
        similarities.label_similarity = lmatrix
        similarities.unique_labels = slabels
    return similarities




from __future__ import print_function
import time
import os

import requests

import pandas as pd

# ------------- Local database loading ----------------------------------

class LocalDBs(object):
    def __init__(self,
                 kegg_db,
                 lm_df,
                 hmdb2kegg_dict,
                 lipidmaps2kegg_dict,
                 kegg2lipidmaps_dict):

        self.kegg_db = kegg_db
        self.lm_df = lm_df
        self.hmdb2kegg_dict = hmdb2kegg_dict
        self.lipidmaps2kegg_dict = lipidmaps2kegg_dict
        self.kegg2lipidmaps_dict = kegg2lipidmaps_dict
        self.new_kegg_records = []

_local_dbs = None
_DB_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), 'dbs'))
        
def load_local_dbs():
    print ('\nLoading local DBs')
    
    kegg_df_fname = os.path.join(_DB_DIR, 'kegg_db.txt')
    kegg_db = load_local_kegg_db(kegg_df_fname)
    
    lm_db_fname = (os.path.join(_DB_DIR, 'lm_metadata.txt'))
    lm_df = load_local_lipidmaps_db(lm_db_fname)    
    
    trans_hmdb2kegg_fname = os.path.join(_DB_DIR, 'trans_hmdb2kegg.txt')
    hmdb2kegg_dict = get_trans_id_table(trans_hmdb2kegg_fname)

    lipidmaps2kegg_dict = lm_df['KEGG_ID'].dropna().to_dict()

    kegg2lipidmaps_dict = {}
    for k, v in lipidmaps2kegg_dict.items():
        kegg2lipidmaps_dict[v] = k
    
    return LocalDBs(kegg_db, lm_df, hmdb2kegg_dict,
                    lipidmaps2kegg_dict, kegg2lipidmaps_dict)


def load_local_kegg_db(db_fname):
    """Load Kegg compound records from a local text DB."""
    
    kegg = {}
    db = open(db_fname)
    curr_c = None
    curr_txt = None
    for line in db:
        if len(line.strip()) == 0:
            continue
        if line.startswith('---- Compound:'):
            if curr_c is not None:
                whole_rec = ''.join(curr_txt)
                kegg[curr_c] = whole_rec
            curr_c = line.split(':')[1].strip()
            curr_txt = []
        elif curr_c is not None:
            curr_txt.append(line)
    whole_rec = ''.join(curr_txt)
    kegg[curr_c] = whole_rec
    db.close()
    return kegg


def load_local_lipidmaps_db(db_fname):
    """Load LIPIDMAPS records from a local text DB."""
    df = pd.read_table(db_fname, index_col=0, dtype={'CHEBI_ID':str})
    return df
    
# ------------- Code related to Kegg compound records -------------------

def get_kegg_section(k_record, sname, whole_section=False):
    """Get the section with name _sname_ from a kegg compound record.
    
       Returns a str, possibly empty."""
    
    in_section = False
    section = []
    
    for line in k_record.splitlines():
        if line.startswith(sname):
            in_section = True
            section.append(line)
        elif in_section and line.startswith(' '):
            section.append(line)
        elif in_section and not line.startswith(' '):
            break

    if whole_section:
        sectionlines = section
    else:
        sectionlines = [line[12:] for line in section]
    return '\n'.join(sectionlines)

def request_kegg_record(c_id, localdict, local_dbs):
    if c_id in localdict:
        return localdict[c_id]
    else:
        record = requests.get('http://rest.kegg.jp/get/' + c_id).text
        outrecord = '---- Compound: {}\n'.format(c_id) + record
        local_dbs.new_kegg_records.append(outrecord)
        return record

# Load ID translation tables as dicts
# IMPORTANT: Use local files, (fetched by fetch_dbs.py)
def get_trans_id_table(fname):
    d = {}
    with open(fname) as f:
        # hmdb:HMDB00002 \t cpd:C00986 \t equivalent
        for line in f:
            if len(line) == 0:
                continue
            foreign, cpd, equiv = line.split('\t')
            foreign = foreign.split(':')[1].strip()
            cpd = cpd.split(':')[1].strip()
            d[foreign] = cpd
    return d


class Progress(object):
    """A progress reporter of % done for a maximum (int) of self.total"""
    def __init__(self, total=1):
        self.reset(total)
    def tick(self):
        self.count +=1
        print(str(round((1-((self.total-self.count)/self.total))*100)) + "% done")
    def reset(self, total=1):
        self.total = total
        self.count = 0


def classes_from_brite(krecord, brite_blacklist=None, trace=False):
    """Parses classes from the BRITE section of a kegg record."""
    classes = [[], [], [], []]

    # replace word BRITE by spaces
    krecord = ' '* 5 + krecord[5:]

    # split by 11 spaces (and discard ":" ), remove trailing \n
    cstrings = [c.rstrip() for c in krecord.split('           ')[1:]]

    brite_id = ''
    for x in cstrings:
##         print('*******************************')
##         print('|{}|'.format(x))
##         print('*******************************')
        
        if x.startswith(' ') and not x.startswith('  '):
            if '[BR:' in x:
                brite_id = x.split('[BR:')[1].split(']')[0]
            elif '[br' in x:
                brite_id = 'br' + x.split('[br')[1].split(']')[0]
            else:
                brite_id = 'br?????'
                
            classes[0].append(x[1:])
        elif x.startswith('  ') and not x.startswith('   '):
            classes[1].append(x[2:])
        elif x.startswith('   ') and not x.startswith('    '):
            classes[2].append(x[3:])
        elif x.startswith('    ') and not x.startswith('     '):
            classes[3].append(x[4:])
    if brite_blacklist is not None:
        if brite_id in brite_blacklist:
            classes = [['null'], [brite_id], [], []]

    classes = ['#'.join(c) for c in classes]
    if trace:
        print ('from BRITE (with id {}):'.format(brite_id))
        print(tuple(classes))
    return tuple(classes)


def classes_from_lipidmaps(lm_id, lm_df, trace=False):
    if trace:
        print('LIPIDMAPS id: {}'.format(lm_id))
##     f = requests.get('http://www.lipidmaps.org/rest/compound/lm_id/' + lm_id + '/all/json').text
##     s = json.loads(f)
##     if f == '[]':
##         return 'null', 'null', 'null', 'null'
##     mm = 'Lipids [LM]'
##     cc = s['core'] if s['core'] is not None else 'null'
##     ss = s['main_class'] if s['main_class'] is not None else 'null'
##     tt = s['sub_class'] if s['sub_class'] is not None else 'null'
    if lm_id not in lm_df.index:
        cc = ['Lipids [LM]', 'null', 'null', 'null']
    else:
        record = lm_df.loc[lm_id]
        cc = ['Lipids [LM]']
        for field in 'CATEGORY', 'MAIN_CLASS', 'SUB_CLASS':
            cc.append(record[field] if pd.notnull(record[field]) else 'null')
    a = tuple(cc)
    if trace:
        print(a)
    return a 


def annotate_compound(compound_id, c_counter, already_fetched, 
                      trace=False, local_dbs=None, brite_blacklist=None):
    c_id = None
    lm_id = None
    hmdb_id = None
    ks_id = None
    
    brite = ''
    dblinks = ''
    in_plants = ''
    
    trans_kegg = compound_id
    trans_lipidmaps = compound_id
    
    if compound_id.startswith('C'):
        c_id = compound_id
        if c_id in local_dbs.kegg2lipidmaps_dict:
            lm_id = local_dbs.kegg2lipidmaps_dict[c_id]
            trans_lipidmaps = lm_id
    
    elif compound_id.startswith('LM'):
        lm_id = compound_id
        if lm_id in local_dbs.lipidmaps2kegg_dict:
            c_id = local_dbs.lipidmaps2kegg_dict[lm_id]
            trans_kegg = c_id
    
    elif compound_id.startswith('HMDB'):
        hmdb_id = compound_id
        if hmdb_id in local_dbs.hmdb2kegg_dict:
            c_id = local_dbs.hmdb2kegg_dict[hmdb_id]
            trans_kegg = c_id
            if c_id in local_dbs.kegg2lipidmaps_dict:
                lm_id = local_dbs.kegg2lipidmaps_dict[c_id]
                trans_lipidmaps = lm_id

    else:
        pass # does MASSTrix use any other DBs for IDs?
    
    if trace:
        print('\n---- compound: {} ({}, {})'.format(compound_id, 
                                             trans_kegg, 
                                             trans_lipidmaps))
    
    if c_id in already_fetched or lm_id in already_fetched:
        if trace:
            print('already looked up')
        return (trans_kegg, trans_lipidmaps, '', '', '', '', '')
        
    if c_id is not None:
        krecord = request_kegg_record(c_id, local_dbs.kegg_db, local_dbs)
        if trace:
            print('(look up {})'.format(c_id))
        c_counter.inc_looks()
        already_fetched.append(c_id)
    
        brite = get_kegg_section(krecord, 'BRITE', whole_section=True)
        dblinks = get_kegg_section(krecord, 'DBLINKS')
        
        if len(dblinks) > 0:
            # find if there are LIPIMAPS or KNApSAcK Xrefs in DBLINKS section
            if 'LIPIDMAPS:' in dblinks:
                for line in dblinks.splitlines():
                    if 'LIPIDMAPS:' in line:
                        lm_id = line.split(':')[1].strip()
                        trans_lipidmaps = lm_id
            if 'KNApSAcK' in dblinks:
                for line in dblinks.splitlines():
                    if 'KNApSAcK:' in line:
                        ks_id = line.split(':')[1].strip()
    
    # get compound taxonomy
    
    # first from BRITE section
    if (c_id is not None) and len(brite) > 0:
        classes = classes_from_brite(brite, 
                                     brite_blacklist=brite_blacklist,
                                     trace=trace)
        if classes[0] == 'null':
            if trace:
                print('BRITE class {} in blacklist. Skipped'.format(classes[1]))
            classes = ('', '', '', '')
    
    # then from LIPIDMAPS
    elif (lm_id is not None):
        if trace:
            print('(look up {})'.format(lm_id))
        classes = classes_from_lipidmaps(lm_id,
                                         local_dbs.lm_df,
                                         trace=trace)
        c_counter.inc_looks()
        already_fetched.append(lm_id)

    else:
        classes = ('', '', '', '')
        if trace:
            print ('No BRITE, no LIPIDMAPS')
    
    # find if present in plants
    in_plants = ''
    if ks_id is not None:
        r = requests.post('http://kanaya.naist.jp/knapsack_jsp/information.jsp?sname=C_ID&word=' + ks_id)
        c_counter.inc_looks()
        if 'Plantae' in r.text:
            in_plants = 'Plantae'
            if trace:
                print(in_plants)
    
    result = [trans_kegg, trans_lipidmaps]
    result.extend(list(classes))
    result.append(in_plants)
    return tuple(result)

class _count_compounds(object):
    def __init__(self):
        self.reset()
    def reset(self):
        self.count = 0
        self.looks_count = 0
    def inc(self, n=1):
        self.count += n
    def inc_looks(self, n=1):
        self.looks_count += n
    def get_count(self):
        return self.count
    def get_looks_count(self):
        return self.looks_count


def annotate_all(peak, c_counter, progress=None, trace=False,
                 local_dbs=None, brite_blacklist=None):
    """Create Pandas Series with compound class annotations."""
    
    data = [[], [], [], [], [], [], []]
    
    if trace:
        print('\n++++++++ PEAK +++++++++++++')
    
    already_fetched = []
    
    for compound_id in peak.split('#'):
        c_counter.inc()
        
        d_compound = annotate_compound(compound_id, 
                                       c_counter, 
                                       already_fetched, 
                                       trace, local_dbs,
                                       brite_blacklist)
        
        for d, i in zip(data, d_compound):
            if len(i) > 0: # don't include empty strings
                d.append(i)
    
    if progress is not None:
        progress.tick()
    
    compressed_data = []
    for i, d in enumerate(data):
        if i < 2:
            compressed_data.append(d)
        else:
            compressed_data.append(set(d))
    
    hash_data = ['#'.join(d) for d in compressed_data]    

    col_names = ['trans_KEGG_Ids',
                 'trans_LipidMaps_Ids',
                 'Major Class', 
                 'Class', 
                 'Secondary Class', 
                 'Tertiary Class',
                 'KNApSAcK']
    
    
    if trace:
        print('\nDATA:')
        for n, d in zip(col_names, hash_data):
            print('{:25s} : {}'.format(n, d))
    
    return pd.Series(hash_data, index=col_names)


def insert_taxonomy(df, brite_blacklist=None, trace=False):
    """Apply annotate_all to kegg or LIPIDMAPS ids."""
    
    progress = Progress(total=len(df['raw_mass']))
    c_counter = _count_compounds()
    local_dbs = load_local_dbs()
    
    if brite_blacklist is not None:
        with open(brite_blacklist) as f:
            brite_blacklist = f.read().splitlines()
    
    df = pd.concat([df, df['KEGG_cid'].apply(annotate_all, 
                                             args=(c_counter, progress,
                                             trace,
                                             local_dbs,
                                             brite_blacklist))], axis=1)
    
    frmt = '\nDone! {} ids processed. {} DB lookups'.format
    print(frmt(c_counter.get_count(), c_counter.get_looks_count()))
    print('updating local Kegg DB')
    kegg_df_fname = os.path.join(_DB_DIR, 'kegg_db.txt')
    with open(kegg_df_fname, 'a') as kf:
        kf.write('\n')
        kf.write('\n'.join(local_dbs.new_kegg_records))
    print('\nDone')
    return df

if __name__ == '__main__':

    print ('\nStarting...\n')
    from metabolinks import read_MassTRIX

    start_time = time.time()

    testfile_name = 'data/MassTRIX_output.tsv'

    results = read_MassTRIX(testfile_name)
    print ("File {} was read\n".format(testfile_name))

    # Clean up uniqueId and "isotope" cols
    results = results.cleanup_cols()

    # check result
    results.info()
    assert len(results.columns) == 12 # there are 24 - 12 = 12 columns
    assert len(results.index) == 15 # there are still 15 peaks

    print ('Starting annotations...')

    # Call the main driver function.
    results = insert_taxonomy(results,
                              #brite_blacklist='../example_data/blacklist.txt',
                              trace=True)

    elapsed_time = time.time() - start_time
    m, s = divmod(elapsed_time, 60)
    print ("Finished in " + "%02dm%02ds" % (m, s))
    print('------------------------------------')

    # results
    results.info()
    # print(results.head())
    
    # Export the annotated dataframe into a MS-Excel file
    # Name it with the same name as the .tsv, replacing tail with '_raw.xlsx'
    
    out_fname = testfile_name[:-4]+'_comptaxa.xlsx'
    results.to_excel(out_fname, header=True, index=False)
    print ("File {} written".format(out_fname))



from __future__ import print_function, absolute_import

from collections import Counter

from six import StringIO, string_types, integer_types

import time

from numpy import nan
import pandas as pd

from metabolinks.taxonomy import insert_taxonomy
from metabolinks.elementcomp import element_composition

"""Functions related to the analysis of MassTRIX results files."""

# ----------------------------------------
# MassTRIXResults
# ----------------------------------------

class MassTRIXResults(pd.DataFrame):
    """A subclass of Pandas DataFrame representing MassTRIX identifications.
    
       Created to facilitate method chaining, and sparing the name space.
       Instances are created mainly by reading from file functions.
    """
    @property
    def _constructor(self):
        return MassTRIXResults    
    
##     def __init__(self, other):
##         # Copy attributes only if other is of good type
##         if isinstance(other, pd.DataFrame):
##             self.__dict__  = other.__dict__.copy()

    def cleanup_cols(self, **kwargs):
        return cleanup_cols(self, **kwargs)
    
    def unfold(self, **kwargs):
        return unfold(self, **kwargs)
    
    def insert_taxonomy(self, *args, **kwargs): 
        return insert_taxonomy(self, *args, **kwargs)
    
    def compute_element_composition(self, *args, **kwargs):
        return element_composition(self, column='KEGG_formula', **kwargs)

# ----------------------------------------
# I/O functions
# ----------------------------------------

def read_MassTRIX(fname, unfolded=False):
    """Reads a MassTRIX file into a Pandas DataFrame object.
       
       On the process, the last line is moved to the beginning and
       is read as the header."""
    
    # store lines in a list
    with open(fname) as f:
        lines = [line.strip() for line in f]
    
    if not unfolded:
        # move the last line to the beginning
        moved_list = [lines[-1]]
        moved_list.extend(lines[:-1]) # last line is not included
        lines = moved_list

    # create a MassTRIXResults, which is also a Pandas DataFrame,
    # reading from the list of strings in memory
    
    mem_string = StringIO('\n'.join(lines))
    
    df = pd.read_table(mem_string)
    df = MassTRIXResults(df)
    return df

# ----------------------------------------
# Unfold functions
#
# these relate to the two MassTRIX formats:
#
# - compact: one peak per line, several putative compounds
#   are separated by #
#
# - unfolded: one putative compound per line
# 
# ----------------------------------------

def _clean_value(v):
    if not isinstance(v, string_types):
        return v
    try:
        return float(v)
    except:
        pass
    return v.strip('').strip(';')


def unfold(df):
    """Unfold a MassTRIX DataFrame.
    
       "Unfolding" means splitting each "peak" line into several
       "compound" lines. # is the split separator.
       
       The number of Ids found in
       column "KEGG_cid" is used as the number of putative compounds in
       each peak.
    """
    
    # split by #, if possible, using str accessor of Pandas Series.
    for c in df.columns:
        try:
            df[c] = df[c].str.split('#')
        except AttributeError:
            pass
    
    # initialize a dict of empty lists, with column names as keys
    unfold_dict = {c:[] for c in df.columns}
    
    # unfold each row of the data frame.
    # if a column is of type list, extend the appropriate lists
    # otherwise repeat elements using row['KEGG_cid'] as the number of repeats
    
    for indx, row in df.iterrows():
        n = len(row['KEGG_cid'])
    
        for label, v in row.iteritems():
            if isinstance(v, list):
                if len(v) < n:
                    v.extend([nan]*(n-len(v)))
                unfold_dict[label].extend(v)
            else:
                unfold_dict[label].extend([v] * n)
    
    # clean values (try to convert to float and remove leading ;)
    for k in unfold_dict:
        unfold_dict[k] = [_clean_value(v) for v in unfold_dict[k]]
    
    return MassTRIXResults(unfold_dict, columns=df.columns)

# ----------------------------------------
# Clean up functions
# ----------------------------------------

def cleanup_cols(df, isotopes=True, uniqueID=True, columns=None):
    """Removes the 'uniqueID' and the 'isotope presence' columns."""
    col_names = []
    if uniqueID:
        col_names.append('uniqueID')
    if isotopes:
        iso_names = ('C13','O18','N15', 'S34', 'Mg25', 'Mg26', 'Fe54',
                     'Fe57', 'Ca44', 'Cl37', 'K41')
        col_names.extend(iso_names)
    if columns is not None:
        col_names.extend(columns)
    #df.drop(col_names, axis=1, inplace=True)
    return df.drop(col_names, axis=1)
    

# ----------------------------------------
# Testing code
# ----------------------------------------

if __name__ == '__main__':

    #testfile_name = '../example_data/masses.annotated.reformat.tsv'
    testfile_name = 'data/MassTRIX_output.tsv'

    results = read_MassTRIX(testfile_name).cleanup_cols()
    
    print("File {} was read\n".format(testfile_name))
    
    results.info()

    print('\n+++++++++++++++++++++++++++++++')
    print('Element compositions:')
    
    compositions = ['CHO', 'CHOS', 'CHON', 'CHONS', 
                    'CHOP', 'CHONP', 'CHONSP']
    
    elem_comp = results.compute_element_composition(compositions=compositions)
    
    for c in compositions + ['other']:
        print(c, elem_comp[c])
    
    print('\n+++++++++++++++++++++++++++++++')

    results = results.unfold()

    print('Unfolded dataframe:\n')
    print('is of type', type(results))
    results.info()

    print('---------------------')
    print(results.head(10))
    
    



from __future__ import print_function
import requests

print('Fetching ID translation table HMDB->Kegg from Kegg...', end=' ')
fname = 'trans_hmdb2kegg.txt'
h = requests.post('http://rest.genome.jp/link/compound/hmdb').text
with open(fname, 'w') as f:
    f.write(h)
print('Done\nFile {} created.\n'.format(fname))

print('Fetching ID translation table LIPIDMAPS->Kegg from Kegg...', end=' ')
fname = 'trans_lipidmaps2kegg.txt'
h = requests.get('http://rest.genome.jp/link/compound/lipidmaps').text
with open(fname, 'w') as f:
    f.write(h)
print('Done\nFile {} created.\n'.format(fname))




from collections import Counter

def element_composition(df, column=None,
                            compositions = ('CHO', 'CHOS',
                                                'CHON', 'CHONS',
                                                'CHOP', 'CHONP',
                                                'CHONSP'),
                            verbose=True):
    
    formulae = df[column].str.split('#').apply(set).apply(list)
    # print(formulas)

    # remove unambigous formulae
    is_1 = [True if len(f) == 1 else False for f in formulae.values]
    formulae = formulae[is_1]
    
    #print(formulae)
    
    # convert to list of strings and remove duplicates
    formulae = list(set([f[0] for f in formulae.values]))
    #print(formulae)
    if verbose:
        print(len(formulae), 'formulae')

    # Calculate element compositions
    comps = []
    for formula in formulae:
        # remove numbers
        exclude = "0123456789,.[]() "
        for chr in exclude:
            formula = formula.replace(chr,'')
        
        # count according to composition groups
        for c in compositions:
            if set(formula) == set(c):
                comps.append(c)
                break
        else:
            comps.append('other')
    elem_comp = Counter(comps)
    
##     for f, c in zip(formulae, comps):
##         print(f, c)
        
    return elem_comp

if __name__ == '__main__':

    from metabolinks.masstrix import read_MassTRIX
    testfile_name = '../example_data/masses.annotated.reformat.tsv'

    df = read_MassTRIX(testfile_name).cleanup_cols()
    
    print("File {} was read\n".format(testfile_name))
    
    df.info()

    print('\n+++++++++++++++++++++++++++++++')
    compositions = ['CHO', 'CHOS', 'CHON', 'CHONS', 
                    'CHOP', 'CHONP', 'CHONSP']
    
    elem_comp = element_composition(df, column='KEGG_formula',
                                    compositions=compositions)
    for c in compositions + ['other']:
        print(c, elem_comp[c])




"""
Extract and build local DBs with relevant information.

Prerequisites:

- From LIPIDMAPS, file LMSDFAll.sdf, obtained (and renamed) from

http://www.lipidmaps.org/resources/downloads/index.html
LMSD Structure-data file (SDF)-> extract and expand LMSDDownload<date>All.sdf.

this file is used to generate lm_metadata.txt (tab delimited).


- From Human Metabolomics Data Base, file hmdb_metabolites.zip, obtained from
http://www.hmdb.ca/downloads -> All metabolites data set.

this file is used to generate trans_hmdb2kegg.txt.

"""

import xml.etree.ElementTree as ET
import zipfile
from six import StringIO

# Extract information from metadata of SDF files of LIPIDMAPS

def get_lm_records(lmfilename):
    """Generator to yield on LM record at a time."""
    with open(lmfilename) as lmfile:
        curr_record = []
        for line in lmfile:
            line = line.strip()
            if line == "$$$$": # record separator
                result = '\n'.join(curr_record)
                result = result.strip()
                if len(result) > 0:
                    yield result
                curr_record = []
            else:
                curr_record.append(line)

def get_lm_metafields(record):
    result = {}
    metas = record.split('M  END', 1)[1]
    split_metas = metas.split('> <')
    for s in split_metas:
        s = s.strip()
        if len(s) == 0:
            continue
        #print(s)
        key, value = s.split('>', 1)
        result[key.strip()] = value.strip()
    return result
        
all_lipids_file = 'metabo_dbs/LMSDFAll.sdf'

out_lm_file = 'dbs/lm_metadata.txt'

fields = ('CATEGORY', 'FORMULA', 'HMDBID', 'CHEBI_ID', 'KEGG_ID',
          'MAIN_CLASS', 'SUB_CLASS', 'SYSTEMATIC_NAME')

print ('processing {} ...'.format(all_lipids_file))

with open(out_lm_file, 'w') as outfile:
    count = 0
    print('LM_ID\t'+'\t'.join(fields), file=outfile)
    
    for record in get_lm_records(all_lipids_file):
        metadata = get_lm_metafields(record)
        lm_id = metadata['LM_ID']
        print(lm_id)
        line = [lm_id]
        
        for field in fields:
            if field in metadata:
                line.append(metadata[field])
            else:
                line.append('')
        
        print('\t'.join(line), file=outfile)
        count +=1
    
print('{} records processed in {}'.format(count, all_lipids_file))
print('File {} generated'.format(out_lm_file))

# Extract information from HMDB
#expname = 'HMDB00010.xml'
all_mets_file = 'metabo_dbs/hmdb_metabolites.zip'
print('\n------------------------------------------\n')
print ('Creating conversions HMDB->KeGGId...')

conv = {}

with zipfile.ZipFile(all_mets_file) as zip:
    for name in zip.namelist():
        if name.startswith('hmdb_metabolites'):
            continue
        record = zip.read(name).decode('utf8')
        
        tree = ET.parse(StringIO(record))
        root = tree.getroot()
        
        accession_tag = root.find('accession')
        acc = accession_tag.text.strip()
        
        kegg_tag = root.find('kegg_id')
        if not kegg_tag.text is None:
            kegg_id = kegg_tag.text.strip()
            if not kegg_id.startswith('.'):
                conv[acc] = kegg_id
                print('{} -> {}'.format(acc, kegg_id))

n_mets = len(zip.namelist())
n_trans = len(conv)
print('{} compounds in HMDB. {} have a KeGGId'.format(n_mets, n_trans))

out_filename = 'dbs/trans_hmdb2kegg.txt'

# try to follow KeGG format for Id translation.
line_frmt = 'hmdb:{}\tcpd:{}\tequivalent'.format

with open(out_filename, 'w') as outfile:
    for hmdb in sorted(list(conv.keys())):
        print(line_frmt(hmdb, conv[hmdb]), file=outfile)

print(out_filename, 'was created')




# -*- coding: utf-8 -*-
"""`metabolinks` is on `Github`_.

.. _github: https://github.com/aeferreira/metabolinks

"""
from io import open
from setuptools import setup, find_packages

def read_file(filename, encoding='utf-8'):
    with open(filename, encoding=encoding) as f:
        content = f.read()
    return content

###############################################################################

CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Science/Research',
    'Intended Audience :: Education',
    'Natural Language :: English',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    "License :: OSI Approved :: MIT License",
    "Programming Language :: Python",
    "Programming Language :: Python :: 2",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Topic :: Scientific/Engineering :: Bio-Informatics",
    "Topic :: Scientific/Engineering :: Chemistry",
    "Topic :: Scientific/Engineering :: Information Analysis"
]

if __name__ == "__main__":
    setup(
        name="metabolinks",
        license="MIT",
        url='https://github.com/aeferreira/metabolinks',
        download_url='https://github.com/aeferreira/metabolinks',
        version='0.51',
        zip_safe=False,
        author='António Ferreira and Gil Pires',
        author_email="aeferreira@fc.ul.pt and gilpires071997@gmail.com",
        maintainer='António Ferreira',
        maintainer_email="aeferreira@fc.ul.pt",
        classifiers=CLASSIFIERS,
        keywords=['Metabolomics', 'Mass Spectrometry',
                  'Data Analysis', 'Ultra-high resolution MS'],
        description="A set of tools for high-resolution MS metabolomics data analysis",
        long_description=read_file('README.rst'),
        packages=['metabolinks'],
        package_data={'metabolinks': [
            'examples/peak_alignment_xcel.ipynb',
            'examples/taxonomy_annotation_example.ipynb',
            'data/data_to_align.xlsx',
            'data/MassTRIX_output.tsv'
        ]},
        include_package_data=True,
        install_requires=['six', 'requests', 'numpy', 
                          'pandas', 'xlrd', 'xlsxwriter',
                          'pytest', 'matplotlib>=2.0'],
    )
        



"""Classes representing MS peak lists.

There are two types of peak lists:

- single (Spectrum)
- aligned (AlignedSpectra). AlignedSpectra represents
multiple peak lists that share the same m/z values (or other peak labels).

Simple data handling functions are implemented as well as I/O to CSV (TSV).

Much more data handling procedures are available through the property `data`
which is a Pandas DataFrame.

"""

from __future__ import print_function

from collections import OrderedDict
import six

import numpy as np
import pandas as pd

from metabolinks.utils import _is_string
from metabolinks.similarity import compute_similarity_measures

class Spectrum(object):
    """A single peak list.

    The underlying data (property `data`) is implemented in a Pandas DataFrame.
    The index holds m/z values or any other peak labels.
    There should be only one column, which holds (ususally) intensity data.

    Attributes
    ----------
    sample_name : str
        The name of this sample.
    label : str
        Optional sample label, used in classification tasks.

    """

    def __init__(self, df, sample_name=None, label=None):
        self._df = df.copy()
        
        self.label = label
        
        self.sample_name = None
        if sample_name is not None:
            self.sample_name = sample_name
        else:
            # read sample name from the name of the only column
            if self._df is not None:
                self.sample_name = self._df.columns[0]
    
    @property
    def data(self):
        """The Pandas DataFrame holding the MS data."""
        return self._df

    def __str__(self):
        res = ['Sample: ' + self.sample_name]
        res.append('Label: {}'.format(self.label))
        res.append('{} peaks'.format(len(self.data)))
        res.append(str(self.data))
        return '\n'.join(res)

    @property
    def mz(self):
        """Get m/z values as a numpy array"""
        res = self._df.dropna().index.values
        return res

    @property
    def all_mz(self):
        """Get m/z values as a numpy array, even for missing data"""
        res = self._df.index.values
        return res

    def fillna(self, value):
        """Get new Spectrum with missing values replaced for a given value."""
        new_df = self._df.fillna(value)
        return Spectrum(new_df, sample_name=self.sample_name, label=self.label)

    def to_csv(self, filename, mz_name=None, sep=None, **kwargs):
        header = True
        if mz_name is not None:
            header = [mz_name] + list(self._df.columns)
        if sep is None:
            sep = '\t'
        self._df.to_csv(filename, header=header, sep=sep, index=True, **kwargs)
    
    def mz_to_csv(self, filename, mz_name=None, **kwargs):
        outdf = self.data.copy()
        outdf.index.name = mz_name
        outdf.to_csv(filename, columns=[])


class AlignedSpectra(object):
    """A set of peak lists, sharing m/z values.

    The underlying data (property `data`) is implemented in a Pandas DataFrame.
    The first column holds m/z values and the other columns hold (ususally)
    intensity data.

    Attributes
    ----------
    sample_names : list of str
        The names of the samples.
    labels : list of str
        Optional sample labels, used in classification tasks.

    """

    def __init__(self, df, sample_names=None, labels=None):
        self._df = df.copy()
        self.labels = labels
        
        if sample_names is None:
            self.sample_names = list(self._df.columns)
        elif isinstance(sample_names, int):
            # read sample_names from columns
            self.sample_names = list(self._df.columns[:sample_names])
        else:
            self.sample_names = sample_names

    @property
    def data(self):
        """The Pandas DataFrame holding the MS data."""
        return self._df

    @property
    def sample_count(self):
        """Get the number of samples."""
        return len(self.sample_names)

    @property
    def mz(self):
        """Get m/z values as a numpy array"""
        res = self.data.index.values
        return res

    def __str__(self):
        return '\n'.join(
               ('{} samples:'.format(self.sample_count),
               str(self.sample_names),
               'Labels: {}'.format(self.labels),
               'Size: {}'.format(len(self.data)),
               str(self.data))
               ) # No, I will not become a js freak


    def label_of(self,sample):
        """Get label from sample name"""
        if self.labels is None:
            return None
        for s, lbl in zip(self.sample_names, self.labels):
            if s == sample:
                return lbl
        return None

    def to_csv(self, filename, header_func=None, sep=None, 
               no_meta_columns=True, **kwargs):
        if sep is None:
            sep = '\t'
        out_df = self._df.copy()
        if no_meta_columns:
            out_df = out_df.iloc[:, :len(self.sample_names)]
        if header_func is None:
            out_df.to_csv(filename, header=True, sep=sep, index=True, **kwargs)
        else:
            # prepend output with result of header_func
            needs_to_close = False
            if _is_string(filename):
                of = open(filename, 'w') 
                needs_to_close = True
            else:
                of = filename

            header = header_func(self) + '\n'
            of.write(header)
            out_df.to_csv(of, header=False, sep=sep, 
                          index=True, **kwargs)

            if needs_to_close:
                of.close()

    def fillna(self, value):
        """Substitute missing values for a given value."""
        return AlignedSpectra(self.data.fillna(value),
                              sample_names=self.sample_names,
                              labels=self.labels)

    def sample(self, sample):
        """Get data for a given sample name, as a Spectrum."""
        df = self.data[[sample]].dropna()
        return Spectrum(df, sample_name=sample, label=self.label_of(sample))

    def unfold(self):
        """Return a list of Spectrum objects, unfolding the peak lists."""
        return [self.sample(name) for name in self.sample_names]

    def label(self, label):
        """Get data for a given label, as an AlignedSpectra object."""
        # build a list grouping samples with a given label
        lcolumns = []
        for c, l in zip(self.data.columns, self.labels):
            if l == label:
                lcolumns.append(c)
        df = self.data[lcolumns].dropna(how='all', subset=lcolumns)
        return AlignedSpectra(df, sample_names=lcolumns,
                                  labels=[label]*len(lcolumns))
    
    def rep_at_least(self, minimum=1):
        df = self._df.copy()
        # build list of unique labels
        unique_labels = list(set(self.labels))
        for label in unique_labels:
            # build a list grouping samples with each given label
            lcolumns = []
            for c, l in zip(self.data.columns, self.labels):
                if l == label:
                    lcolumns.append(c)
            n_label = len(lcolumns)
            lessthanmin = df[lcolumns].count(axis=1) < minimum
            df.loc[lessthanmin, lcolumns] = np.nan
        df = df.dropna(how='all')
        return AlignedSpectra(df, sample_names=self.sample_names,
                                  labels=self.labels)

    def common_label_mz(self, label1, label2):
        mz1 = self.label(label1).mz
        mz2 = self.label(label2).mz
        u = np.intersect1d(mz1, mz2)
        return u
    
    def exclusive_label_mz(self, label):
        # build list of unique other labels
        slabels = []
        for lbl in self.labels:
            if lbl not in slabels:
                slabels.append(lbl)
        slabels = [lbl for lbl in slabels if lbl != label]

        remaining = self.label(label).mz
        for lbl in slabels:
            remaining = np.setdiff1d(remaining, self.label(lbl).mz)
        return remaining


def read_spectrum(filename, label=None):
    s = pd.read_table(filename, index_col=False)
    # keep only the first two columns
    s = s.iloc[:, [0,1]]
    s = s.set_index(s.columns[0])
    return Spectrum(s, label=label)


def read_aligned_spectra(filename, labels=None):
    s = pd.read_table(filename, index_col=False)
    s = s.set_index(s.columns[0])
    return AlignedSpectra(s, labels=labels)


def read_spectra_from_xcel(file_name,
                           sample_names=None,
                           labels=None,
                           header_row=1,
                           common_mz= False,
                           verbose=True):

    spectra_table = OrderedDict()

    wb = pd.ExcelFile(file_name).book
    header = header_row - 1

    if verbose:
        print ('------ Reading MS-Excel file -------\n{}'.format(file_name))

    for sheetname in wb.sheet_names():
        if verbose:
            print ('- reading sheet "{}"...'.format(sheetname))

        # if sample_names argument if present (not None) then
        # if an integer, read row with sample names,
        # otherwise sample_names must a list of names for samples
        snames = []
        if sample_names is not None:
            if isinstance(sample_names, six.integer_types):
                sh = wb.sheet_by_name(sheetname)
                snames = sh.row_values(sample_names - 1)
                snames = [s for s in snames if len(s.strip()) > 0]
                header = sample_names
            else:
                snames = sample_names

        # read data (and discard empty xl columns)
        df = pd.read_excel(file_name,
                           sheetname=sheetname,
                           header=header)
        df = df.dropna(axis=1, how='all')

##         df.info()
##         print('============================================')
##         print(df.head())

        # if sample names are not set then
        # use "2nd columns" headers as sample names
        # if common_mz then use headers from position 1
        if len(snames) > 0:
            sample_names = snames
        else:
            if common_mz:
                sample_names = df.columns[1:]
            else:
                sample_names = df.columns[1::2]

        # split in groups of two (each group is a spectrum)
        results = []
        if not common_mz:
            j = 0
            for i in range(0, len(df.columns), 2):
                spectrum = df.iloc[:, i: i+2]
                spectrum.index = range(len(spectrum))
                spectrum = spectrum.dropna()
                spectrum = spectrum.set_index(spectrum.columns[0])
                spectrum = Spectrum(spectrum, sample_name=sample_names[j])
                results.append(spectrum)

                j = j + 1
        else:
            j = 0
            for i in range(1, len(df.columns)):
                spectrum = df.iloc[:, [0, i]]
                spectrum.index = range(len(spectrum))
                spectrum = spectrum.dropna()
                spectrum = spectrum.set_index(spectrum.columns[0])
                spectrum = Spectrum(spectrum, sample_name=sample_names[j])
                results.append(spectrum)

                j = j + 1

        if labels is not None:
            for i, spectrum in enumerate(results):
                spectrum.label = labels[i]

        if verbose:
            for spectrum in results:
                name = spectrum.sample_name
                print ('{:5d} peaks in sample {}'.format(spectrum.data.shape[0], name))
        spectra_table[sheetname] = results

    return spectra_table



def _sample_data():
    return """m/z	s38	s39	s40	s32	s33	s34
97.58868	1073218	1049440	1058971	2351567	1909877	2197036
97.59001	637830	534900	582966	1440216	1124346	1421899
97.59185	460092	486631		1137139	926038	1176756
97.72992			506345			439583
98.34894	2232032	2165052	1966283			
98.35078	3255288	2813578	2516386			
98.35122		2499163	2164976			
98.36001			1270764	1463557	1390574	
98.57354				4627491	6142759	
98.57382		3721991	3338506			4208438
98.57497	6229543	3347404	2327096			
98.57528				2510001	1989197	4377331
98.57599	6897403	3946118				
98.57621			3242232	2467520		4314818
98.57692	8116811	5708658	3899578			
98.57712				2418202	986128	4946201
98.57790	3891025	3990442	3888258	2133404		3643682
98.57899		1877649	1864650	1573559		1829208
99.28772	2038979				3476845	"""


if __name__ == '__main__':
    from six import StringIO

    sample = StringIO(_sample_data())
   
    print('Reading one spectrum (from aligned reads only first) ------------')
    spectrum = read_spectrum(sample)
    print(spectrum,'\n')
    spectrum.data.info()

    print('\nm/z values (excluding missing data) ----------')
    print(spectrum.mz)

    print('\nSpectrum with missing values filled with zeros ----------')
    spectrumzero = spectrum.fillna(0)
    print(spectrumzero.data)

    print('\nSaving  mz into a file ----------')
    mzfile = StringIO()    
    spectrum.mz_to_csv(mzfile, mz_name='moverz')
    print('\n--- Resulting file:')
    print(mzfile.getvalue())

    print('\nReading aligned spectra (all of them) ----------')
    labels=['v1', 'v1', 'v1', 'v2', 'v2', 'v2']
    sample.seek(0)
    spectra = read_aligned_spectra(sample, labels=labels)
    print(spectra,'\n')
    spectra.data.info()

    print('\nData of sample s38 ----------')
    print(spectra.sample('s38'))
    
    print('\nm/z of sample s38 ----------')
    print(spectra.sample('s38').mz)
    
    print('\nData of label v1 ----------')
    print(spectra.label('v1'))
    
    print('\nm/z of label v1 ----------')
    print(spectra.label('v1').mz)

    print('\nFiltered fewer than 2 per label')
    print('\n-- Original\n')
    print(spectra)
    print('\n-- With a minimum of 2 replicates\n')
    print(spectra.rep_at_least(minimum=2))

    print('\nSpectra with missing values filled with zeros ----------')
    spectrazero = spectra.fillna(0)
    print(spectrazero.data)

    print('\nUnfolding spectra ----------')
    unfolded = spectra.unfold()
    
    for u in unfolded:
        print(u)
        print('+++++')
    
    print('\nSaving aligned spectra into a file ----------')
    def header(s):
        # this returns a header suitable for various metabolomics tools
        line1 = 'Labels,{}'.format(','.join(s.labels))
        line2 = 'Samples,'+','.join(s.sample_names)
        return '\n'.join([line1, line2])

    samplefile = StringIO()    
    spectra.to_csv(samplefile, header_func=header, sep=',')
    print('\n--- Resulting file:')
    print(samplefile.getvalue())
    #spectra.to_csv('testout.txt', header_func=header, sep=',')

    print('\nCommon m/z between labels (v1,v2) ----------')
    print(spectra.common_label_mz('v1', 'v2'))

    print('\nm/z values exclusive to label v1 ----------')
    print(spectra.exclusive_label_mz('v1'))

    print('\nComputing similarity measures ----------')
    
    sim = compute_similarity_measures(spectra)
    
    print('\n- Sample similarity --')
    print(sim.sample_similarity)
        
    print('\n- Label similarity --')
    print(sim.label_similarity)
    
